import re
import traceback
import unittest
from copy import copy
from pprint import pprint
from random import random
from typing import List, Dict

from unvibe import ai_call, magic_entities
from unvibe.tests_container import TestsContainer
from unvibe.config import config, config_get_or
from unvibe.log import log
from unvibe.magic import cleanup_implementation, MagicEntity
from unvibe.rand import up_to_1
from unvibe.state import State
from unvibe.ui import create_page_and_open_browser


def generate_new_state(count, state: State, temperature: float, tests_container: TestsContainer) -> State:
    """Generates a new state for program space search"""
    new_state = State()
    new_state.mes = state.mes
    new_state.count = count
    new_state.tests = state.tests
    new_state.temperature = temperature
    first_error = [state.errors[0]] if len(state.errors) > 0 else []
    prompt, resp_text = ai_call(state.mes, state.context, state.tests, first_error, temperature)
    new_state.prompt = prompt
    new_state.ai_output = resp_text
    impls = parse_ai_output(resp_text)
    new_state.impls = impls
    pprint(impls)
    # If returned implementations is subset of expected implementations:
    has_all_impls = {me.name for me in state.mes} <= set(impls.keys())

    if has_all_impls:
        log(f'Received {len(impls)} implementations, expected {len(state.mes)}')
        for me in new_state.mes:
            if me.name in impls:
                cleaned_up = cleanup_implementation(impls[me.name], me.__class__)
                me.set_impl(cleaned_up)
            else:
                raise Exception(f'Expected implementation for {me.name}')
        new_state.context = new_state.build_context_from_magic_entities()
        run_tests(tests_container, new_state)
        # Reset implementations:
        for me in new_state.mes:
            me.set_impl(None)
    else:
        log('LLM OUTPUT:', resp_text)
        new_state.score = 0
        expected_func_names = ', '.join([me.name for me in new_state.mes])
        new_state.errors = [f'Expected implementations for {expected_func_names}.\n'
                            f'Did you use one <implement> for each function? Try again.']
        log(new_state.errors)
    return new_state


def start_search(mes: List[MagicEntity], tests_container: TestsContainer, sources='', display_tree=False):
    log('Using model', config['ai']['model'])

    # Check all magic entities are registered
    for me in mes:
        assert me in magic_entities, f'{me} not registered with @ai'

    # Run the tree search
    root, states = search(mes, tests_container, sources, display_tree)
    best_state = states[0]
    if display_tree:
        file = create_page_and_open_browser(root)
        print(f'Created execution report {file}')
    return best_state


def get_temperatures(depth):
    random_spread = config_get_or('search', 'random_spread', 2)
    if depth == 0:
        random_spread = config_get_or('search', 'initial_spread', 10)
    random_type = config_get_or('search', 'random_type', 'uniform')
    max_temperature = config_get_or('search', 'max_temperature', 0.7)

    # Generate a bunch of rand temperatures, but always try temp=0
    if random_type == 'increasing':
        return [up_to_1.pop(0) for _ in range(random_spread)]
    elif random_type == 'uniform':
        return [0] + [random() * max_temperature for _ in range(random_spread)]
    else:
        raise Exception(f'Unknown random_type "{random_type}". Use either "increasing" or "uniform"')


def build_initial_context(mes, sources):
    if len(sources) > 0:
        # We have a list of source files, so we replace each MagicEntity with its implementation
        new_context = copy(sources)
        for me in mes:
            if me.impl is None:
                continue
            idx = sources.find(me.orig_code)
            if idx < 0:
                log(f'Could not find {me.orig_code} in context')
                continue
            else:
                new_context = new_context.replace(me.orig_code, me.impl)
        return new_context
    else:
        # We don't have files, just MEs, so we concatenate the implementations
        new_context = ''
        for me in mes:
            new_context += me.orig_code + '\n'
        return new_context


def search(mes: List[MagicEntity], test_container: TestsContainer, sources='', display_tree=False):
    take_best_n = config_get_or('search', 'take_best_n', 3)
    max_depth = config_get_or('search', 'max_depth', 10)

    # Generate the root state
    root = State()
    root.count = 0
    root.mes = mes
    root.temperature = 0.0
    root.tests = test_container.get_source()
    assert type(root.tests) == str
    root.errors = []
    root.score = -1  # root state has no score
    root.context = build_initial_context(mes, sources)
    root.orig_context = root.context
    states = [root]

    found = False
    count = 0
    for depth in range(max_depth):
        log('Depth', depth)
        new_states = []
        # For each state, generate a bunch of new states feeding back the current test errors
        for state in states:
            # Generate a bunch of new states: generate code with LLMs and run the tests to get the score
            for temp in get_temperatures(depth):
                log('Temperature', temp)
                count += 1
                new_state = generate_new_state(count, state, temp, test_container)
                if new_state is None:
                    continue
                state.children.append(new_state)
                new_states.append(new_state)
                if new_state.score == 1:
                    # Early quit
                    log('Found perfect score')
                    found = True
                    break
            if found: break
        states = states + new_states
        states = sorted(states, key=lambda s: random())
        states.sort(key=lambda s: s.score, reverse=True)
        log('Scores   ', [s for s in states], 'Picking the best', take_best_n)
        states = states[:take_best_n]
        log('Selected ', [s for s in states])
        if display_tree:
            create_page_and_open_browser(root)
        if found: break
    return root, states


def run_tests(tests_container: TestsContainer, new_state: State):
    """Runs the tests and saves the score/assertions info/errors in the new_state"""
    test_suite = tests_container.generate_test_suite()
    runner = unittest.TextTestRunner()
    result = runner.run(test_suite)
    if result.testsRun == 0:
        raise Exception(f"Test class {test_suite} has no tests.")
    errors_count = len(result.failures) + len(result.errors)
    try:
        if not test_suite.ai_test_case:
            raise Exception('Not using unvibe.TestCase')
        total_passed_assertions = test_suite.total_passed_assertions
        total_executed_assertions = test_suite.total_executed_assertions
        total_failed_assertions = test_suite.total_failed_assertions
        log('Using unvibe.TestCase')
        total_assertions = tests_container.count_assertions()
        log(f"Total assertions: {total_assertions}, Passed: {total_passed_assertions}, "
            f"Executed: {total_executed_assertions}, Failed: {total_failed_assertions}")
        if total_assertions == 0 or total_passed_assertions > total_assertions:
            raise Exception('Invalid assertions count')
        score = total_passed_assertions / total_assertions
    except Exception as exc:
        log(exc)
        score = 1 - errors_count / result.testsRun
        total_passed_assertions, total_executed_assertions, total_assertions, total_failed_assertions = None, None, None, None
    error_strings = []
    for test_suite, error_str in result.errors + result.failures:
        error_strings.append(cleanup_error_str(error_str))
    new_state.passed_assertions = total_passed_assertions
    new_state.executed_assertions = total_executed_assertions
    new_state.failed_assertions = total_failed_assertions
    new_state.total_assertions = total_assertions
    new_state.errors = error_strings
    new_state.score = score


def remove_lines_with(lines, is_target, minus=1, plus=2):
    try:
        target_line = -1
        for i, line in enumerate(lines):
            if is_target(line):
                target_line = i
                break
        if target_line >= 0:
            # Remove the target line and those before and after:
            lines = lines[:target_line - minus] + lines[target_line + plus:]
        return lines
    except Exception as e:
        traceback.print_exc()
        return lines


def cleanup_error_str(error_str):
    # Remove every reference to the file path:
    error_str = re.sub(r'File ".*", line', 'line', error_str)
    # Find the line with "eval" and "self.func_name":
    lines = error_str.split('\n')

    lines = remove_lines_with(lines, lambda line: 'eval(' in line and '{self.func_name}' in line)
    lines = remove_lines_with(lines, lambda line: 'exec(code)' in line)
    lines = remove_lines_with(lines, lambda line: 'return ___eval(' in line)
    lines = remove_lines_with(lines, lambda line: 'line ' in line and 'in wrapper' in line, minus=0, plus=2)
    lines = remove_lines_with(lines, lambda line: 'line ' in line and 'in wrapper' in line, minus=0, plus=2)

    # Put back together the redacted lines
    error_str = '\n'.join(lines)
    return error_str


def parse_ai_output(t: str) -> Dict:
    # find anything between <implements name="..."> and </implement>
    found = re.findall(r'<implement name="(.+?)">(.*?)</implement>', t, re.DOTALL)
    found_dict = dict(found)
    return found_dict


def index_of_first_non_empty_char(line: str) -> int:
    for i, c in enumerate(line):
        if c != ' ':
            return i
    return -1


def remove_extra_indentation(code: str):
    lines = code.split('\n')
    for line in lines:
        i = index_of_first_non_empty_char(line)
        if i >= 0:
            break
    ret = []
    for line in lines:
        ret.append(line[i:])
    return '\n'.join(ret)



import ollama
import anthropic
from typing import List
from openai import OpenAI
from google import genai

from . import MagicFunction
from .config import config
from .disk_cache import disk_cached

total_input_tokens = 0
total_output_tokens = 0

cached = disk_cached


@cached
def call_gemini(system, prompt, temperature):
    client = genai.Client(api_key=config['ai']['api_key'])
    resp = client.models.generate_content(
        model=config['ai']['model'],
        contents=system + '\n' + prompt,
    )
    return resp.text


@cached
def call_openai(system, prompt, temperature):
    # TODO: plugin temperature
    client = OpenAI(api_key=config['ai']['api_key'], base_url=config['ai']['base_url'])
    resp = client.chat.completions.create(
        model="deepseek-chat",
        messages=[
            {"role": "system", "content": system},
            {"role": "user", "content": prompt},
        ],
        max_tokens=config['ai']['max_tokens'],
        # From deepseek.com:
        # USE CASE                  TEMPERATURE
        # Coding / Math:	                0.0
        # Data Cleaning / Data Analysis:    1.0
        # General Conversation:	            1.3
        # Translation:                      1.3
        # Creative Writing / Poetry:        1.5
        temperature=temperature,
        stream=False
    )
    # print('llm resp.choices[0].message.content:', resp.choices[0].message.content)
    return resp.choices[0].message.content


@cached
def call_claude(system, prompt, temperature):
    client = anthropic.Anthropic(api_key=config['ai']['api_key'])
    message = client.messages.create(
        model=config['ai']['model'],
        max_tokens=1000,
        temperature=temperature,
        messages=[{"role": "user", "content": [{"type": "text", "text": prompt + '\n' + system}]}]
    )
    # print('llm output:', message.content)
    global total_input_tokens, total_output_tokens
    total_input_tokens += message.usage.input_tokens
    total_output_tokens += message.usage.output_tokens
    input_tokens_cost_1M, output_tokens_cost_1M = 0.80, 4.00
    cost = (total_input_tokens * input_tokens_cost_1M + total_output_tokens * output_tokens_cost_1M) / 1_000_000
    print(f'Total input tokens: {total_input_tokens}, Total output tokens: {total_output_tokens}, Est. Cost: {cost}')
    return message.content[0].text


@cached
def call_ollama(system, prompt, temperature, model):
    client = ollama.Client(host=config['ai']['host'])
    resp = client.generate(
        model=model,
        prompt='\n' + prompt + '\n' + system,
        options=dict(temperature=temperature))
    return resp.response


example = '''
<input>
def pluck(l, key):
    """Pluck a key from a list of dictionaries."""
    pass
    
assert pluck([{'name': 'Alice'}, {'name': 'Bob'}], 'name') == ['Alice', 'Bob']
</input>
<implement name="pluck">
def pluck(l, key):
    return [d[key] for d in l]
</implement>

<input>
def permutations_str(s: str):
    """Return all permutations of a string."""
    pass

class TestPermutations(unittest.TestCase):
    def test_permutations(self):
        self.assertEqual(set(permutations_str('abc')), {'abc', 'acb', 'bac', 'bca', 'cab', 'cba'})
</input>
<implement name="permutations_str">
def permutations_str(s: str):
    from itertools import permutations
    return set(''.join(p) for p in permutations(s))
</implement>

<input>

class Utils:
        def is_palindrome(s: str):
            return s == s[::-2]
    
class TestIsPalindrome(unittest.TestCase):
    def test_is_palindrome(self):
        utils = Utils()
        self.assertTrue(utils.is_palindrome('racecar'))
        self.assertFalse(utils.is_palindrome('hello'))
<errors>
Traceback (most recent call last):
    line 19, in test_is_palindrome
        self.assertTrue(utils.is_palindrome('racecar'))
</errors>
</input>

<implement name="is_palindrome">
        def is_palindrome(s: str):
            return s == s[:-1]
</implement>
'''

system = ("- You only write code inside the <implement> tags.\n"
          "- No explanations before or after.\n"
          "- Implement ONLY the requested functions. "
          "- Use one <implement> for each function.\n"
          "- If errors and implementation are already in the input, implement again fixing the errors.\n"
          "- Use as many <implement> as needed. Every function should be in it own <implement> tag.\n"
          "- Write all import inside functions.\n"
          "- If you encounter ModuleNotFoundError, try to make do without that module.\n"
          "- Don't write unit-tests. Don't change existing tests.\n")


def ai_call(mfs: List[MagicFunction], context, tests, errors, temperature) -> str:
    assert context.strip() != '', 'Context should not be empty'  # TODO: Catch earlier
    func_names_str = ', '.join([mf.name for mf in mfs])
    errors_tag = ''
    fix_msg = ''
    if len(errors) > 0:
        errors_tag = '<errors>\n' + '\n'.join(errors) + '</errors>'
        fix_msg = 'Fix the errors! you have already tried many times, must try something else.'
    prompt = f'''
{example}
<input>
{context}

{tests}

{errors_tag}
</input>

Implement the functions: {func_names_str}
{fix_msg}
'''
    provider = config['ai']['provider']
    if provider == 'claude':
        resp_text = call_claude(system, prompt, temperature)
    elif provider == 'openai':
        resp_text = call_openai(system, prompt, temperature)
    elif provider == 'gemini':
        resp_text = call_gemini(system, prompt, temperature)
    elif provider == 'ollama':
        resp_text = call_ollama(system, prompt, temperature, model=config['ai']['model'])
    else:
        raise NotImplementedError(f'{provider} not implemented')
    return system + prompt, resp_text



import unittest

from unittest import TestSuite


class CountingTestSuite(TestSuite):
    """A unittest.TestSuite that counts the number of successful assertions."""

    def __init__(self, tests=()):
        self.total_passed_assertions = 0
        self.total_executed_assertions = 0
        self.total_failed_assertions = 0
        self.ai_test_case = False
        super().__init__(tests)

    def _tearDownPreviousClass(self, test, result):
        if hasattr(result, 'passed_assertions'):
            self.ai_test_case = True
            if hasattr(result, 'passed_assertions'):
                self.total_passed_assertions = result.passed_assertions
            if hasattr(result, 'executed_assertions'):
                self.total_executed_assertions = result.executed_assertions
            if hasattr(result, 'failed_assertions'):
                self.total_failed_assertions = result.failed_assertions
        super()._tearDownPreviousClass(test, result)


class TestCase(unittest.TestCase):
    """
    Wraps all assertEqual, assertTrue etc. methods from unittest.TestCase,
    but counting the successful ones in a global counter
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for name in dir(self):
            if name.startswith('assert'):
                method = getattr(self, name)
                setattr(self, name, self._wrap_assert(method))

    def incr_assertions_counter(self, name):
        if hasattr(self._outcome.result, name):
            setattr(self._outcome.result, name, getattr(self._outcome.result, name) + 1)
        else:
            setattr(self._outcome.result, name, 1)

    def _wrap_assert(self, method):
        def wrapper(*args, **kwargs):
            self.incr_assertions_counter('executed_assertions')
            # with self.subTest():
            try:
                method(*args, **kwargs)
                self.incr_assertions_counter('passed_assertions')
                # print('\t\t', self._outcome.result.passed_assertions)
            except AssertionError as e:
                self.incr_assertions_counter('failed_assertions')
                raise e

        return wrapper



import argparse
import sys
from datetime import datetime
from pathlib import Path

from unvibe import magic_entities, log, project_name
from unvibe.tests_container import FolderPatternTestsContainer
from unvibe.core import start_search
from unvibe.state import State

epilog = f'''examples:
  {project_name} src/ tests/                    # Uses every file in src/ and tests/ folders
  {project_name} src/main.py test/test_main.py  # Use only the indicated files
'''


def parse_args_and_run_main() -> State:
    parser = argparse.ArgumentParser(
        description='UnitAI - Generate code that pass unit-tests',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog)
    parser.add_argument('sources', help='sources folder')
    parser.add_argument('tests', help='unit-tests folder')
    parser.add_argument('-o', '--output_folder', default='.', help='output folder for new implementations')
    parser.add_argument('-p', '--pattern', default='test*.py', help='pattern for test files')
    parser.add_argument('-d', '--display_report', help='Display HTML report for the search tree', action='store_true', default=False)
    if len(sys.argv) <= 2:
        parser.print_help()
        sys.exit(0)
    args = parser.parse_args()
    return main(args)


def main(args) -> (State, str):
    log('Sources:', args.sources)
    log('Tests:', args.tests)
    tests_container = FolderPatternTestsContainer(args.tests, args.pattern)
    tests_container.generate_test_suite()
    sources = get_sources_context(args)
    best_state = start_search(magic_entities, tests_container, sources, args.display_report)
    output_file = write_output_folder(best_state, args.output_folder)
    return best_state, output_file


def get_sources_context(args) -> str:
    sources = Path(args.sources)
    if sources.is_file():
        return sources.read_text()
    elif sources.is_dir():
        context = ''
        for file in sources.glob('*.py'):
            context += file.read_text() + '\n'
        return context
    else:
        raise FileNotFoundError(f'File or folder not found: {args.sources}')


def write_output_folder(state: State, output_folder):
    if state.score == 1:
        msg = '# This implementation passed all tests'
    elif state.score == 0:
        msg = ('# Could not find any implementation that passes any test. Check the errors returned from the tests'
               '# You may want to try a different model, change max_temperature, max_depth or random_spread params.')
    else:
        msg = '# This is the best implementation we found, but it did not pass all the tests. Check the errors below.'

    test_case_msg = f'(make your test classes extend {project_name}.TestCase to see this)' if state.passed_assertions is None else ''
    final_text = f'''# {project_name.capitalize()} Execution output.
{msg}
# Score: {state.score}
# Passed assertions: {state.passed_assertions}/{state.total_assertions} {test_case_msg}
'''
    if len(state.errors) > 0:
        all_errors = '=====\n'.join(state.errors)
        commented_errors = ''
        for line in all_errors.split('\n'):
            commented_errors += '# ' + line + '\n'
        final_text += commented_errors
    for key, impl in state.impls.items():
        final_text += f'\n{impl}\n'
    time = datetime.now().strftime('%H-%M-%S')
    Path(output_folder).mkdir(exist_ok=True)
    file_path = Path(output_folder) / f'{project_name}_output_{time}'
    with open(file_path, 'w') as f:
        f.write(final_text)
    print('Written results to', file_path)
    return file_path


if __name__ == '__main__':
    parse_args_and_run_main()



import json
import os
import sys
from bigtree import Node, print_tree, tree_to_dot

from unvibe import project_name
from unvibe.core import State


def create_page_and_open_browser(root):
    page_file = create_html_page(root)
    if sys.platform == 'darwin':
        os.system(f'open {page_file}')
    elif sys.platform == 'win32':
        os.system(f'start {page_file}')
    elif sys.platform == 'linux':
        os.system(f'xdg-open {page_file}')
    return page_file


def create_html_page(root):
    template_path = os.path.join(os.path.dirname(__file__), 'ui.html')
    with open(template_path, 'r') as f:
        template = f.read()
    tree_html = build_html_tree(root)
    try:
        img_file = save_bigtree_img(root)
        template = template.replace('__TREE_IMG__', img_file)
    except Exception as exc:
        pass
    template = template.replace('__TREE__', tree_html)
    template = template.replace('__JSON_TREE__', json.dumps(root.to_dict()))
    func_names = '_'.join([mf.name for mf in root.mes])
    report_file = f'{project_name}_report_{func_names}.html'  # Use the same file for each run
    with open(report_file, 'w') as f:
        f.write(template)
    return report_file


def build_html_tree(root):
    html = '<ul>'
    for child in root.children:
        html += f'<li><span class="node" id="{child.count}">{str(child)}</span>'
        if child.children:
            html += build_html_tree(child)
        html += '</li>'
    html += '</ul>'
    return html


def make_bigtree(node: State, is_root=True):
    if is_root:
        name = ','.join([mf.name for mf in node.mes])
    else:
        name = str(node)
    bt = Node(name)
    for child in node.children:
        bt >> make_bigtree(child, is_root=False)
    return bt


def save_bigtree_img(root: State):
    bt = make_bigtree(root)
    func_names = ','.join([mf.name for mf in root.mes])
    print_tree(bt)
    graph = tree_to_dot(bt, node_colour="gold")
    file_name = f"tree_{func_names}.png"
    graph.write_png(file_name)
    return file_name